@@ -45,13 +45,19 @@ struct _WnckScreenPrivate
   GList *stacked_windows;
   /* in 0-to-N order */
   GList *workspaces;
-
+  
   WnckWindow *active_window;
   WnckWorkspace *active_workspace;
 
   Pixmap bg_pixmap;
   
   guint update_handler;  
+
+#ifdef HAVE_STARTUP_NOTIFICATION
+  SnDisplay *sn_display;
+#endif
+  
+  guint showing_desktop : 1;
   
   /* if you add flags, be sure to set them
    * when we create the screen so we get an initial update
@@ -62,6 +68,7 @@ struct _WnckScreenPrivate
   guint need_update_active_window : 1;
   guint need_update_workspace_names : 1;
   guint need_update_bg_pixmap : 1;
+  guint need_update_showing_desktop : 1;
 };
 
 enum {
@@ -75,6 +82,7 @@ enum {
   APPLICATION_OPENED,
   APPLICATION_CLOSED,
   BACKGROUND_CHANGED,
+  SHOWING_DESKTOP_CHANGED,
   LAST_SIGNAL
 };
 
@@ -87,6 +95,7 @@ static void update_workspace_list   (WnckScreen      *screen);
 static void update_active_workspace (WnckScreen      *screen);
 static void update_active_window    (WnckScreen      *screen);
 static void update_workspace_names  (WnckScreen      *screen);
+static void update_showing_desktop  (WnckScreen      *screen);
 
 static void queue_update            (WnckScreen      *screen);
 static void unqueue_update          (WnckScreen      *screen);              
@@ -108,6 +117,7 @@ static void emit_application_opened       (WnckScreen      *screen,
 static void emit_application_closed       (WnckScreen      *screen,
                                            WnckApplication *app);
 static void emit_background_changed       (WnckScreen      *screen);
+static void emit_showing_desktop_changed  (WnckScreen      *screen);
 
 static gpointer parent_class;
 static guint signals[LAST_SIGNAL] = { 0 };
@@ -249,6 +259,19 @@ wnck_screen_class_init (WnckScreenClass *klass)
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
+  signals[SHOWING_DESKTOP_CHANGED] =
+    g_signal_new ("showing_desktop_changed",
+                  G_OBJECT_CLASS_TYPE (object_class),
+                  G_SIGNAL_RUN_LAST,
+#if 0
+                  /* FIXME when we can break ABI add this */
+                  G_STRUCT_OFFSET (WnckScreenClass, showing_desktop_changed),
+#else
+                  0,
+#endif
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);
 }
 
 static void
@@ -269,12 +292,32 @@ wnck_screen_finalize (GObject *object)
   g_list_free (screen->priv->workspaces);
 
   screens[screen->priv->number] = NULL;
+
+#ifdef HAVE_STARTUP_NOTIFICATION
+  sn_display_unref (screen->priv->sn_display);
+#endif
   
   g_free (screen->priv);
   
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
+#ifdef HAVE_STARTUP_NOTIFICATION
+static void
+sn_error_trap_push (SnDisplay *display,
+                    Display   *xdisplay)
+{
+  gdk_error_trap_push ();
+}
+
+static void
+sn_error_trap_pop (SnDisplay *display,
+                   Display   *xdisplay)
+{
+  gdk_error_trap_pop ();
+}
+#endif /* HAVE_STARTUP_NOTIFICATION */
+
 static void
 wnck_screen_construct (WnckScreen *screen,
                        int         number)
@@ -284,6 +327,12 @@ wnck_screen_construct (WnckScreen *screen,
   screen->priv->xscreen = ScreenOfDisplay (gdk_display, number);
   screen->priv->number = number;
 
+#ifdef HAVE_STARTUP_NOTIFICATION
+  screen->priv->sn_display = sn_display_new (gdk_display,
+                                             sn_error_trap_push,
+                                             sn_error_trap_pop);
+#endif
+  
   screen->priv->bg_pixmap = None;
   
   _wnck_select_input (screen->priv->xroot,
@@ -298,6 +347,7 @@ wnck_screen_construct (WnckScreen *screen,
   screen->priv->need_update_active_window = TRUE;
   screen->priv->need_update_workspace_names = TRUE;
   screen->priv->need_update_bg_pixmap = TRUE;
+  screen->priv->need_update_showing_desktop = TRUE;
   
   queue_update (screen);
 }
@@ -333,6 +383,28 @@ wnck_screen_get (int index)
   return screens[index];
 }
 
+WnckScreen*
+_wnck_screen_get_existing (int number)
+{
+  g_return_val_if_fail (gdk_display != NULL, NULL);
+  g_return_val_if_fail (number < ScreenCount (gdk_display), NULL);
+
+  if (screens != NULL)
+    return screens[number];
+  else
+    return NULL;
+}
+
+WnckScreen*
+wnck_screen_get_default (void)
+{
+  int default_screen;
+
+  default_screen = DefaultScreen (gdk_display);
+
+  return wnck_screen_get (default_screen);
+}
+
 /**
  * wnck_screen_get_for_root:
  * @root_window_id: an Xlib window ID
@@ -363,6 +435,33 @@ wnck_screen_get_for_root (gulong root_window_id)
   return NULL;
 }
 
+/**
+ * wnck_screen_get_workspace:
+ * @screen: a #WnckScreen
+ * @number: a workspace index
+ * 
+ * Gets the workspace numbered @number for screen @screen, or returns %NULL if
+ * no such workspace exists.
+ * 
+ * Return value: the workspace, or %NULL
+ **/
+WnckWorkspace*
+wnck_screen_get_workspace (WnckScreen *screen,
+			   int         number)
+{
+  GList *list;
+  
+  /* We trust this function with property-provided numbers, it
+   * must reliably return NULL on bad data
+   */
+  list = g_list_nth (screen->priv->workspaces, number);
+
+  if (list == NULL)
+    return NULL;
+  
+  return WNCK_WORKSPACE (list->data);
+}
+
 /**
  * wnck_screen_get_active_workspace:
  * @screen: a #WnckScreen 
@@ -436,6 +535,24 @@ wnck_screen_get_windows_stacked (WnckScreen *screen)
   return screen->priv->stacked_windows;
 }
 
+/**
+ * _wnck_screen_get_gdk_screen:
+ * @screen: a #WnckScreen
+ * 
+ * Obtains the #GdkScreen referring to the same screen
+ * as @screen.
+ * 
+ * Return value: a #GdkScreen
+ **/
+GdkScreen *
+_wnck_screen_get_gdk_screen (WnckScreen *screen)
+{
+  g_return_val_if_fail (WNCK_IS_SCREEN (screen), NULL);
+
+  return gdk_display_get_screen (gdk_display_get_default (),
+				 screen->priv->number);
+}
+
 /**
  * wnck_screen_force_update:
  * @screen: a #WnckScreen
@@ -470,6 +587,38 @@ wnck_screen_get_workspace_count (WnckScreen *screen)
   return g_list_length (screen->priv->workspaces);
 }
 
+/**
+ * wnck_screen_change_workspace_count:
+ * @screen: a #WnckScreen
+ * @count: requested count
+ * 
+ * Asks the window manager to change the number of workspaces.
+ **/
+void
+wnck_screen_change_workspace_count (WnckScreen *screen,
+                                    int         count)
+{
+  XEvent xev;
+  
+  g_return_if_fail (WNCK_IS_SCREEN (screen));
+  g_return_if_fail (count >= 1);
+  
+  xev.xclient.type = ClientMessage;
+  xev.xclient.serial = 0;
+  xev.xclient.window = screen->priv->xroot;
+  xev.xclient.send_event = True;
+  xev.xclient.display = DisplayOfScreen (screen->priv->xscreen);
+  xev.xclient.message_type = _wnck_atom_get ("_NET_NUMBER_OF_DESKTOPS");
+  xev.xclient.format = 32;
+  xev.xclient.data.l[0] = count;
+  
+  XSendEvent (DisplayOfScreen (screen->priv->xscreen),
+              screen->priv->xroot,
+              False,
+              SubstructureRedirectMask | SubstructureNotifyMask,
+              &xev);
+}
+
 void
 _wnck_screen_process_property_notify (WnckScreen *screen,
                                       XEvent     *xevent)
@@ -513,6 +662,12 @@ _wnck_screen_process_property_notify (WnckScreen *screen,
       screen->priv->need_update_bg_pixmap = TRUE;
       queue_update (screen);
     }
+  else if (xevent->xproperty.atom ==
+           _wnck_atom_get ("_NET_SHOWING_DESKTOP"))
+    {
+      screen->priv->need_update_showing_desktop = TRUE;
+      queue_update (screen);
+    }
 }
 
 static gboolean
@@ -920,7 +1075,7 @@ update_workspace_list (WnckScreen *screen)
         {
           WnckWorkspace *space;
 
-          space = _wnck_workspace_create (old_n_spaces + i);
+          space = _wnck_workspace_create (old_n_spaces + i, screen);
 
           screen->priv->workspaces = g_list_append (screen->priv->workspaces,
                                                     space);
@@ -951,7 +1106,7 @@ update_workspace_list (WnckScreen *screen)
       
       tmp = tmp->next;
     }
-
+  
   tmp = created;
   while (tmp != NULL)
     {
@@ -959,14 +1114,16 @@ update_workspace_list (WnckScreen *screen)
       
       tmp = tmp->next;
     }
-
+  g_list_free (created);
+  
   tmp = deleted;
   while (tmp != NULL)
     {
-      _wnck_workspace_destroy (WNCK_WORKSPACE (tmp->data));
+      g_object_unref (tmp->data);
       
       tmp = tmp->next;
     }
+  g_list_free (deleted);
 
   /* Active workspace property may now be interpretable,
    * if it was a number larger than the active count previously
@@ -997,7 +1154,7 @@ update_active_workspace (WnckScreen *screen)
                            &number))
     number = -1;
   
-  space = wnck_workspace_get (number);
+  space = wnck_screen_get_workspace (screen, number);
 
   if (space == screen->priv->active_workspace)
     return;
@@ -1050,21 +1207,24 @@ update_workspace_names (WnckScreen *screen)
                                _wnck_atom_get ("_NET_DESKTOP_NAMES"));
 
   copy = g_list_copy (screen->priv->workspaces);
-
+  
   i = 0;
   tmp = copy;
   while (tmp != NULL)
-    {
+    {      
       if (names && names[i])
-        _wnck_workspace_update_name (tmp->data, names[i]);
+        {
+          _wnck_workspace_update_name (tmp->data, names[i]);
+          ++i;
+        }
       else
         _wnck_workspace_update_name (tmp->data, NULL);
 
-      ++i;
       tmp = tmp->next;
     }
 
   g_strfreev (names);
+
   g_list_free (copy);
 }
 
@@ -1089,6 +1249,26 @@ update_bg_pixmap (WnckScreen *screen)
   emit_background_changed (screen);
 }
 
+static void
+update_showing_desktop (WnckScreen *screen)
+{
+  int showing_desktop;
+  
+  if (!screen->priv->need_update_showing_desktop)
+    return;
+  
+  screen->priv->need_update_showing_desktop = FALSE;
+
+  showing_desktop = FALSE;
+  _wnck_get_cardinal (screen->priv->xroot,
+                      _wnck_atom_get ("_NET_SHOWING_DESKTOP"),
+                      &showing_desktop);
+
+  screen->priv->showing_desktop = showing_desktop != 0;
+  
+  emit_showing_desktop_changed (screen);
+}
+
 static void
 do_update_now (WnckScreen *screen)
 {
@@ -1106,7 +1286,8 @@ do_update_now (WnckScreen *screen)
   update_active_workspace (screen);
   update_active_window (screen);
   update_workspace_names (screen);
-
+  update_showing_desktop (screen);
+  
   update_bg_pixmap (screen);
 }
 
@@ -1229,6 +1410,14 @@ emit_background_changed (WnckScreen *screen)
                  0);
 }
 
+static void
+emit_showing_desktop_changed (WnckScreen *screen)
+{
+  g_signal_emit (G_OBJECT (screen),
+                 signals[SHOWING_DESKTOP_CHANGED],
+                 0);
+}
+
 gboolean
 wnck_screen_net_wm_supports (WnckScreen *screen,
                              const char *atom)
@@ -1259,3 +1448,111 @@ wnck_screen_get_height (WnckScreen *screen)
 
   return gdk_screen_height ();
 }
+
+Screen *
+_wnck_screen_get_xscreen (WnckScreen *screen)
+{
+  return screen->priv->xscreen;
+}
+
+int
+_wnck_screen_get_number (WnckScreen *screen)
+{
+  return screen->priv->number;
+}
+
+int
+wnck_screen_try_set_workspace_layout (WnckScreen *screen,
+                                      int         current_token,
+                                      int         rows,
+                                      int         columns)
+{
+  int retval;
+  
+  g_return_val_if_fail (WNCK_IS_SCREEN (screen),
+                        WNCK_NO_MANAGER_TOKEN);
+  
+  retval = _wnck_try_desktop_layout_manager (screen->priv->xscreen, current_token);
+
+  if (retval != WNCK_NO_MANAGER_TOKEN)
+    {
+      _wnck_set_desktop_layout (screen->priv->xscreen, rows, columns);
+    }
+
+  return retval;
+}
+
+void
+wnck_screen_release_workspace_layout (WnckScreen *screen,
+                                      int         current_token)
+{
+  _wnck_release_desktop_layout_manager (screen->priv->xscreen,
+                                        current_token);
+
+}
+
+gboolean
+wnck_screen_get_showing_desktop (WnckScreen *screen)
+{
+  g_return_val_if_fail (WNCK_IS_SCREEN (screen), FALSE);
+  
+  return screen->priv->showing_desktop;
+}
+
+void
+wnck_screen_toggle_showing_desktop (WnckScreen *screen,
+                                    gboolean    show)
+{
+  g_return_if_fail (WNCK_IS_SCREEN (screen));
+
+  _wnck_toggle_showing_desktop (screen->priv->xscreen,
+                                show);
+}
+
+#ifdef HAVE_STARTUP_NOTIFICATION
+SnDisplay*
+_wnck_screen_get_sn_display (WnckScreen *screen)
+{
+  g_return_val_if_fail (WNCK_IS_SCREEN (screen), NULL);
+  
+  return screen->priv->sn_display;
+}
+#endif /* HAVE_STARTUP_NOTIFICATION */
+
+void
+_wnck_screen_change_workspace_name (WnckScreen *screen,
+                                    int         number,
+                                    const char *name)
+{
+  int n_spaces;
+  char **names;
+  int i;
+  
+  n_spaces = wnck_screen_get_workspace_count (screen);
+
+  names = g_new0 (char*, n_spaces + 1);
+
+  i = 0;
+  while (i < n_spaces)
+    {
+      if (i == number)
+        names[i] = (char*) name;
+      else
+        {
+          WnckWorkspace *workspace;
+          workspace = wnck_screen_get_workspace (screen, i);
+          if (workspace)
+            names[i] = (char*) wnck_workspace_get_name (workspace);
+          else
+            names[i] = (char*) ""; /* maybe this should be a g_warning() */
+        }
+      
+      ++i;
+    }
+
+  _wnck_set_utf8_list (screen->priv->xroot,
+                       _wnck_atom_get ("_NET_DESKTOP_NAMES"),
+                       names);
+
+  g_free (names);
+}
